/* *************************************************************************
                          lightcache.cpp  -  description
                             -------------------
    begin                : Wed Oct 16 2002        
    copyright            : (C) 2002 by Micha Riser
    email                : mriser@gmx.net
 ************************************************************************* */

/* *************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ************************************************************************* */

#include "lightcache.h"
#include "scene.h"
#include "lightsource.h"
#include "prime.h"

/*#include <iostream>
using std::cout;*/

HitStatistics LightCache::Hash::hitstat("LightCache","Evaluations","Hits");
HitStatistics LightCache::Hash::insertstat("LightCache","Inserts","Collisions");

void LightCache::setType(interpolationtype type, int size) {
    
    itype = type;
    map.setSize(size);

}

LightList* LightCache::sample(const Vector3& location, CLR& blockingerror) {

    LightList* l;
    bool succ = map.get(location, blockingerror, l);
    if (!succ) {
        l = scene->getLightList(location, blockingerror);
        map.put(location, blockingerror, l);
    }
    return l;

}

LightList* LightCache::lookup(const Vector3& location, CLR& blockingerror, DBL precision) {

///usage    precision *= 0.1;

/*    if (evals % 1000000 <8) {
        cout << "evals: "<<evals;
        cout << " / media evals: "<<Media3D::evals;
        cout << " / miss :"<<evals-hits;
        cout << " / hitsize: "<<1.0*hitsize/hits;
        cout << "\n";
    }*/

    if (itype == EXACT) {
        LightList* l;
        bool succ = map.get(location,blockingerror,l);
        if (!succ) {
           l = scene->getLightList(location, blockingerror);
           map.put(location, blockingerror, l);
        }
        return l;
    }

    // round precision
    int pp;
    frexp(precision, &pp);
    DBL grids = ldexp(0.5,pp);

    // round location to grid
    Vector3 locmin, diff;
    for(int i=0; i<3; i++) {
        diff.a[i] = fmod(location.a[i],grids);
        locmin.a[i] = location.a[i] - diff.a[i];
        if (location.a[i]<0) {locmin.a[i]-=grids; diff.a[i]=grids+diff.a[i];}
    }

    if (itype == NEAREST || itype == TRILINEAR_BEST) {

        Vector3 nearest = locmin;
        for(int i=0; i<3; i++) if (diff.a[i]>grids/2) nearest.a[i]+=grids;
        
        LightList* l;
        bool succ = map.get(nearest, blockingerror, l);
        if (!succ) { 
            l = scene->getLightList(nearest, blockingerror);
            map.put(nearest, blockingerror, l);
        }
        if (itype == NEAREST) return l;
    }

    LightList* ret = 0;

    CLR err;

    blockingerror = 0;
    CLR min = 100000;
    CLR max = 0;

    CLR weight = 0;
    CLR grids3 = 1.0/grids/grids/grids;
    for (int i=0; i<8; i++) {
  
        Vector3 test = locmin;
        CLR w = 1;
        for(int j=0; j<3; j++) if (i&(1<<j)) {
            test.a[j]+=grids;
            w*=diff.a[j];
        } else {
            w*=(grids-diff.a[j]);
        }

        if (w*grids3 >0.01) {
            LightList* l;
            l = sample(test, err);

            CLR tmp=0;
            LightList* it = l;
            while(it !=0) {
                tmp += it->colour.norm();
                it = it->next;
            }
            if (tmp<min) min = tmp;
            if (tmp>max) max = tmp;
  
            ret = LightList::merge(ret,l,w);
            blockingerror += err*w;
            weight += w;
        }

    }

    blockingerror /= weight;
    blockingerror = 0.25*blockingerror + 4*(max-min);
    LightList* it = ret;

    while(it != 0) {

        it->colour.scale(1/weight);
        it = it->next;

    }

    return ret;

}

LightCache::Hash::Hash() {
    size = 0;
    data = 0;
}

LightCache::Hash::~Hash() {
    for(int i=0; i<size; i++) delete(data[i].list);
    delete[] data;
}

void LightCache::Hash::setSize(int s) {

    for(int i=0; i<size; i++) delete(data[i].list);
    delete[] data;
    
    size = Prime::nextPrimeBiggerThan(s);
    sondmod = Prime::nextPrimeSmallerThan(int(0.98*s));
    data = new MapEntry[size];
    
//    cout << size << " "<<sondmod <<"\n";
    
}

bool LightCache::Hash::get(const Vector3& location, CLR& err, LightList* &ret) {

    unsigned int loc_i = vectorToInt(location);
    unsigned int i = calcHashValue(loc_i);
    unsigned int sond = calcHashValueSond(loc_i);

    for(int j=0; j<sondsize; j++) {
        if (data[i].hitcount != 0 && data[i].location.equals(location) == true) {
            data[i].hitcount += 1<<sondsize;
            err = data[i].blockingerror;
            ret = data[i].list;
            hitstat.evalSuccess();
            return true;
        }
        i = (i+sond)%size;
    }

    hitstat.evalFailed();
    return false;

}



void LightCache::Hash::put(const Vector3& loc, CLR err, LightList* list) {

    insertstat.eval();
    unsigned int lp_i = vectorToInt(loc);
    unsigned int i = calcHashValue(lp_i);
    unsigned int sond = calcHashValueSond(lp_i);

    unsigned int worst = i;
    unsigned int worst_count = 1000000;

    for(int j=0; j<sondsize; j++) {

        if (data[i].hitcount<worst_count) {
            worst = i;
            worst_count = data[i].hitcount;
        }
        data[i].hitcount = data[i].hitcount/2+1;

        if (data[i].hitcount <= 1) {

            delete(data[i].list);
            data[i] = MapEntry(loc, err, list);
            break;

        }

        if (j == sondsize-1) {

            delete(data[worst].list);
            data[worst] = MapEntry(loc, err, list);
            insertstat.success();
            break;

        }

        i = (i+sond)%size;

    }
  
}

unsigned int LightCache::Hash::vectorToInt(const Vector3&v) {
/*
    unsigned int res = 0;
    unsigned int len = sizeof(v);
    const char* p = reinterpret_cast<const char*>(&v);
    while(len--) res = (res<<1)^*p++;
*/
    unsigned int res = 0;
    for(int i=0; i<3; i++) {
        const long long* p = reinterpret_cast<const long long*>(&(v.a[i]));
        unsigned int val = ((*p)>>38) ^ ((*p)>>49);
        res = (res<<4) ^ val;
    }
    return res;
}

unsigned int LightCache::Hash::calcHashValue(unsigned int res) {
    return res % size;
} 

unsigned int LightCache::Hash::calcHashValueSond(unsigned int res) {
    return 1 + (res % sondsize);
}

LightCache::MapEntry::~MapEntry() {}

